#include <algorithm>
#include <cassert>
#include "bg.hpp"

namespace cfg {
using std::find;
using std::any_cast;
using utils::LabelList;
using mir::StmType;
using mir::Jump;
using mir::Cjump;
using std::cout, std::endl;

static unordered_map<utils::Label*, mir::StmList*> env;

BlockGraph::BlockGraph(mir::Block* block): Graph(){
    
    
    for(auto& stmList: *block->stmLists){
        assert(stmList->size() > 0 &&
            stmList->front()->kind == StmType::LABEL);
        env[static_cast<mir::Label*>(stmList->front())->label] = stmList;
        lookup(stmList);
    }
    env[block->label] = new mir::StmList{new mir::Return(new mir::Const(1))};
    exit = lookup(env[block->label]);
    entry = lookup(block->stmLists->front());

    for(auto& stmList: *block->stmLists){
        auto lastStm = stmList->back();
        LabelList* ll;
        switch (lastStm->kind)
        {
        case StmType::JUMP:
        {
            Jump* j = static_cast<Jump*>(lastStm);
            ll = new LabelList{j->label};
        }
        break;
        case StmType::CJUMP:
        {
            Cjump* c = static_cast<Cjump*>(lastStm);
            ll = new LabelList{c->true_label, c->false_label};
        }
        break;
        case StmType::RETURN:
        {
            mir::Return* c = static_cast<mir::Return*>(lastStm);
            ll = new LabelList();
        }
        break;
        default:
        assert(0);
        break;
        }
        
        for(auto& l: *ll){
            if(env.count(l)){
                enter(stmList, env[l]);
            }  
        }
    }
}

Node* BlockGraph::lookup(mir::StmList* b){
    for(auto& node: *mynodes){
        if(any_cast<mir::StmList*>(node->info) == b){
            return node;
        }
    }
    any info = b;
    return new Node(this, info);
}

void BlockGraph::enter(mir::StmList* b1, mir::StmList* b2){
    auto n1 = lookup(b1);
    auto n2 = lookup(b2);
    addEdge(n1, n2);
}
void BlockGraph::show(){
    //cout << "bg has " << nodecount << " nodes" << endl;
    for(auto& n: *mynodes){
        //cout << n->mykey << ": ";
        for(auto& s: *std::any_cast<mir::StmList*>(n->info)){
            s->print();
        }
        for(auto& succ: *n->succs){
            //cout << succ->mykey << " ";
        }
        //cout << endl;
    }
}

};
