#include "cdg.hpp"
namespace ssa{
    void  ControlDependenceGraph::build(){
        std::unordered_map<Node*, Node*> nodemap;   
        if(!exitnode) {
            exitnode = new Node(this, nullptr);
            nodemap[exitnode] = exitnode;
        }else{
            auto newnode = new Node(this, nullptr);
            nodemap[exitnode] = newnode;
        }
        
        // nodelist->emplace_back(entrynode);
        // nodelist->emplace_back(exitnode);
        for(auto node: *nodelist){
            if(node == exitnode) continue;
            Node* newnode = new Node(this, node->info);
            // nodelist->emplace_back(newnode);
            nodemap[node] = newnode;
        }
        entrynode = new Node(this, nullptr);
        for(auto node: *nodelist){
            Node* newnode = nodemap[node];
            for(auto pred: *(node->preds)){
                addEdge(newnode, nodemap[pred]);
            }
            for(auto succ: *(node->succs)){
                addEdge(nodemap[succ], newnode);
            }
        }
        addEdge(nodemap[nodelist->front()], entrynode);
        //TODO: add exit node
        addEdge(nodemap[exitnode], entrynode);
        //show();
        DomiTable postdomi = compute_domi(mynodes);

        for(auto node: *mynodes){
            for(auto pred: *(node->preds)){
                rmEdge(pred, node);
            }
            for(auto succ: *(node->succs)){
                rmEdge(node, succ);
            }
        }
        std::cout << "all edge removed"<<std::endl;
        //using df to build cdg
        for(auto& nodey: *mynodes){
            auto node_df = postdomi[nodey]->dominance_frontier;

            for(auto nodex: node_df){
                addEdge(nodex, nodey);
            }
        }
        std::cout << "cdg added"<<std::endl;
    }

    void ControlDependenceGraph::show(){
        std::cout << "cdg has " << nodecount << " nodes" << std::endl;
        for(auto& n: *mynodes){
            
            std::cout << n->mykey << ": ";
            for(auto& succ: *n->succs){
                std::cout << succ->mykey << " ";
            }
            std::cout << std::endl;
            if(n == entrynode|| n == exitnode) continue;
            // for(auto& s: *std::any_cast<mir::StmList*>(n->info)){
            //     s->print();
            // }
            
            
        }
    }

    class BasicGraph: public cfg::Graph{
    public:
        BasicGraph(){
        };
        void show(){
            std::cout << "basic has " << nodecount << " nodes" << std::endl;
            for(auto& n: *mynodes){
                std::cout << n->mykey << ": ";
                for(auto& succ: *n->succs){
                    std::cout << succ->mykey << " ";
                }
                std::cout << std::endl;
            };
    };
    };
    void test_cdg(){
        BasicGraph* graph = new BasicGraph();
        std::vector<Node*> nodes;
        for(int i = 0; i < 7; i++){
            Node* newnode = new Node(graph, nullptr);
            nodes.push_back(newnode);
        }
        graph->addEdge(nodes[0], nodes[1]);
        graph->addEdge(nodes[1], nodes[2]);
        graph->addEdge(nodes[1], nodes[3]);
        graph->addEdge(nodes[2], nodes[4]);
        graph->addEdge(nodes[2], nodes[5]);
        graph->addEdge(nodes[4], nodes[6]);
        graph->addEdge(nodes[5], nodes[6]);
        graph->addEdge(nodes[6], nodes[1]);
        
        Node* exitnode = new Node(graph, nullptr);
        graph->addEdge(nodes[3], exitnode);
        //graph->show();
        auto cdg = new ControlDependenceGraph(graph->mynodes, exitnode);
    };
}
