#include "de.hpp"
namespace ssa{
bool DeadcodeElimination::has_side_effect(mir::Stm* stm){
    switch (stm->kind) {
        case mir::StmType::MOVE: {
            auto movstm = static_cast<mir::Move*>(stm);
            auto exp = movstm->src;
            switch (exp->kind){
            case mir::ExpType::CALL:{
                return true;
                break;
            }
            default:
                break;
            }
            if (movstm->dst->kind == mir::ExpType::MEM) {
                return true;
            }
            // if (movstm->dst->kind == mir::ExpType::TEMP) {
            //     return false;
            // }
        }
        case mir::StmType::EXP: {
            auto expstm = static_cast<mir::ExpStm*>(stm);
            if (expstm->exp->kind == mir::ExpType::CALL) {
                return true;
            }
            break;
        }
        case mir::StmType::CJUMP: {
            return false;
            break;
        }
        case mir::StmType::RETURN: {
            return true;
            break;
        }
    }
    return false;
}

mir::Stm* get_condition_stm(Node* node){
    mir::StmList* stmlist = std::any_cast<mir::StmList*>(node->info);
    if(!stmlist) return nullptr;
    auto stm = stmlist->back();
    if(stm->kind == mir::StmType::CJUMP){
        return stm;
    }
    return nullptr;
}

void DeadcodeElimination::mask(){
    auto cdgnodes = cdg->mynodes;
    for(auto node: *cdgnodes){
        mir::StmList* stmlist = std::any_cast<mir::StmList*>(node->info);
        for(auto& stm: *stmlist){
            stm_to_block[stm] = node;
            auto def = get_def(stm);
            if(def){
                def_stm[static_cast<mir::Temp*>(def)->temp] = stm;
            }
            if(has_side_effect(stm)){
                active_stm.insert(stm);
                worklist.insert(stm);
                active_block.insert(node);
            }
        }
    }
    while (!worklist.empty())
    {
        auto stm = *worklist.begin();
        worklist.erase(stm);
        auto uses = get_uses(stm);
        for(auto use: uses){
            worklist.insert(def_stm[static_cast<mir::Temp*>(use)->temp]);
            active_stm.insert(def_stm[static_cast<mir::Temp*>(use)->temp]);
            active_block.insert(stm_to_block[stm]);
        }

        //condition branch
        auto this_block = stm_to_block[stm];
        for (auto pred: *this_block->preds){
            auto con_stm = get_condition_stm(pred);
            if(con_stm){
                worklist.insert(con_stm);
                active_stm.insert(con_stm);
                active_block.insert(stm_to_block[con_stm]);
            }
        }
    }
    
}
}//namespace ssa