%option noyywrap yylineno

%{
	#include <cstdlib>
	#include <string>
	#include "../structure/ast/ast.hpp"
	#include "parser.hpp"

	extern int yyleng;

%}


ID								[a-z_A-Z][0-9a-z_A-Z]*
FORMAT							[\"][^\n]*[\"]
INTNUMBER_8						0[0-7]+
INTNUMBER_16					0[xX][0-9a-fA-F]+
INTNUMBER_10					[1-9][0-9]*
SKIP							[ \n\t\r]


FLOAT1							([0-9]*)\.[0-9]*
FLOAT2							[0-9]+[eE][\+\-]?[0-9]+
FLOAT3							[0-9]*\.[0-9]*[eE][\+\-]?[0-9]+
FLOAT16							0[xX][0-9a-fA-F]*[\.]?[0-9a-fA-F]*[pP][\+\-]?[0-9]+


%start COMMENT
%start LINECOMMENT

%%

<INITIAL>"//" {
	BEGIN LINECOMMENT;
}

<LINECOMMENT>\n {
	BEGIN INITIAL;
}

<LINECOMMENT>[^\n]* {}

<INITIAL>"/*" {
	BEGIN COMMENT;
}

<COMMENT>"*/" {
	BEGIN INITIAL;
}

<COMMENT>[^\n\*]* {}
<COMMENT>\* {}

return		{ return RETURN; }
if			{ return IF;}
else 		{ return ELSE; }
while		{ return WHILE; }
break		{ return BREAK; }
continue	{ return CONTINUE; }
void 		{ return VOID; }
int 		{ return INTT; }
float		{ return FLOAT; }
const		{ return CONST; }

starttime"("")"	{ return STARTTIME; }
stoptime"("")"	{ return STOPTIME; }


{FORMAT} {
	yylval.format = new std::string(yytext);
	return FORMAT;
}

{ID} {
	yylval.id = new std::string(yytext);
	return ID;
}


0 {
	yylval.token = 0;
	return INTNUMBER;
}

{INTNUMBER_8} {
	yylval.token = (int)strtol(yytext, (char **)NULL, 8);
	return INTNUMBER;
}

{INTNUMBER_16} {
	yylval.token = (int)strtol(yytext, (char **)NULL, 16);
	return INTNUMBER;
}

{INTNUMBER_10} {
	yylval.token = (int)strtol(yytext, (char **)NULL, 10);
	return INTNUMBER;
}

{SKIP} {}

{FLOAT1} {
	yylval.floatnumber = new std::string(yytext);
	return FLOATNUMBER;
}

{FLOAT2} {
	yylval.floatnumber = new std::string(yytext);
	return FLOATNUMBER;
}

{FLOAT3} {
	/* decimal fps with . */
	// fprintf(stderr,"%s\n", yytext);
	yylval.floatnumber = new std::string(yytext);
	// fprintf(stderr, "%f\n", std::stof(*yylval.floatnumber));
	return FLOATNUMBER;
}

{FLOAT16} {
	// fprintf(stderr,"%s\n", yytext);
	yylval.floatnumber = new std::string(yytext);
	// fprintf(stderr, "%f\n", std::stof(*yylval.floatnumber));
	return FLOATNUMBER;
}

"+"			{ return ADDOP; }
"-"			{ return SUBOP; }
"!"			{ return NOTOP; }
"*"			{ yylval.mul = AST::mul_t::MULT; return MULOP; }
"/"			{ yylval.mul = AST::mul_t::DIV; return MULOP; }
"%"			{ yylval.mul = AST::mul_t::REM; return MULOP; }
"<"			{ yylval.rel = AST::rel_t::LT; return RELOP; }
">"			{ yylval.rel = AST::rel_t::GT; return RELOP; }
"<="		{ yylval.rel = AST::rel_t::LE; return RELOP; }
">="		{ yylval.rel = AST::rel_t::GE; return RELOP; }
"=="		{ yylval.equal = AST::equal_t::EQ; return EQUALOP; }
"!="		{ yylval.equal = AST::equal_t::NE; return EQUALOP; }
"&&"		{ return LANDOP; }
"||"		{ return LOROP; }

"{" { return '{';}
"}" { return '}';}
"(" { return '(';}
")" { return ')';}
"[" { return '[';}
"]" { return ']';}
";" { return ';';}
"=" { return '=';}
"," { return ',';}


. {
	yylval.token = yytext[0];
	std::cerr << yytext << std::endl;
	exit(yytext[0]);
}


%%
