%{
	#include "../structure/ast/ast.hpp"
	#include <iostream>
	#include <cstdlib>
	#include <string>
	#include <cstdio>

	AST::CompUnits *root;
	extern char *yytext;

	extern int yylex();
	extern int yyget_lineno();
	void yyerror(const char *s) { std::cerr << s << std::endl; exit(yytext[0]);}

%}


%union {
	int token;
	std::string *id;
	std::string *floatnumber;
	std::string *format;
	AST::mul_t mul;
	AST::rel_t rel;
	AST::equal_t equal;
	AST::CompUnits *root;
	AST::CompUnit *compunit;
	AST::Decl *decl;
	AST::ConstDecl *constdecl;
	AST::btype_t btype;
	AST::ConstDef *constdef;
	AST::ConstDefs *constdefs;
	AST::VarDecl *vardecl;
	AST::VarDefs *vardefs;
	AST::VarDef *vardef;
	AST::Arr_Index *arr_index;
	AST::Arr_Init *arr_init;
	AST::InitVals *initvals;
	AST::InitVal *initval;
	AST::FuncDef *funcdef;
	AST::Parameters *parameters;
	AST::Parameter *parameter;
	AST::Block *block;
	AST::BlockItems *blockitems;
	AST::BlockItem *blockitem;
	AST::Stmt *stmt;
	AST::Exp *exp;
	AST::IdExp *idexp;
	AST::Exps *exps;
	AST::Number *number;
	AST::Lval *lval;
	AST::unary_op_t unaryop;
}

%token<token> BREAK CONTINUE RETURN WHILE IF ELSE VOID
%token<token> FLOAT CONST INTT 
%token<token> STARTTIME STOPTIME
%token<token> INTNUMBER
%token<token> ADDOP SUBOP  LANDOP LOROP NOTOP
%token<mul> MULOP
%token<equal> EQUALOP
%token<rel> RELOP
%token<floatnumber> FLOATNUMBER
%token<format> FORMAT
%token<id> ID

%type <root> COMPUNITS
%type <compunit> COMPUNIT
%type <decl> DECL
%type <constdecl> CONSTDECL
%type <vardecl> VARDECL
%type <constdefs> CONSTDEFS
%type <constdef> CONSTDEF
%type <vardefs> VARDEFS
%type <vardef> VARDEF
%type <arr_index> ARR_INDEX
%type <initval> INITVAL
%type <arr_init> ARR_INIT
%type <initvals> INITVALS
%type <block> BLOCK
%type <blockitems> BLOCKITEMS
%type <blockitem> BLOCKITEM
%type <funcdef> FUNCDEF
%type <parameters> PARAMETERS
%type <parameter> PARAMETER
%type <stmt> STMT
%type <stmt> IFSTMT WHILESTMT ASSIGNSTMT EXPSTMT BREAKSTMT CONTINUESTMT RETURNSTMT
%type <stmt> STARTTIMESTMT STOPTIMESTMT
%type <exps> EXPS
%type <exp> EXP
%type <exp> PRIMARYEXP UNARYEXP MULEXP ADDEXP RELEXP EQEXP LANDEXP LOREXP 
%type <exp> CALLEXP FORMATEXP
%type <idexp> IDEXP
%type <btype> BTYPE
%type <number> NUMBER
%type <unaryop> UNARYOP
%type <lval> LVAL

%left LOROP
%left LANDOP
%left EQUALOP
%left RELOP
%left ADDOP SUBOP
%left MULOP
%left prec1
%left prec2
%left ELSE

%start COMPUNITS

%%

COMPUNITS: COMPUNITS COMPUNIT {
	/* CompUnits -> CompUnits CompUnit */
	$$ = $1;
	$$->list.push_back($2);

	root = $$;

} | COMPUNIT {
	/* CompUnits -> CompUnit */
	$$ = new AST::CompUnits(yyget_lineno());
	$$->list.push_back($1);
	
	root = $$;
}

COMPUNIT: DECL  {
	/* CompUnit -> Decl */
	$$ = static_cast<AST::CompUnit *>($1);
} | FUNCDEF {
	/* CompUint -> FuncDef */
	$$ = static_cast<AST::CompUnit *>($1);
}

DECL: CONSTDECL {
	/* Decl -> ConstDecl */
	$$ = static_cast<AST::Decl *>($1);
} | VARDECL {
	/* Decl -> VarDecl */
	$$ = static_cast<AST::Decl *>($1);
}


CONSTDECL: CONST BTYPE CONSTDEFS ';' {
	/* ConstDecl -> const Btype ConstDef (, ConstDef)* ; */
	$$ = new AST::ConstDecl($2, $3, yyget_lineno());
}

CONSTDEFS: CONSTDEF {
	$$ = new AST::ConstDefs(yyget_lineno());
	$$->list.push_back($1);
} | CONSTDEFS ',' CONSTDEF {
	$$ = $1;
	$1->list.push_back($3);
}

BTYPE: INTT {
	$$ = AST::btype_t::INT;
} | FLOAT {
	$$ = AST::btype_t::FLOAT;
} | VOID {
	$$ = AST::btype_t::VOID;
}

CONSTDEF: IDEXP ARR_INDEX '=' INITVAL {
	/* ConstDef -> ID ( [Exp] )* = InitVal */
	$$ = new AST::ConstDef($1, $2, $4, yyget_lineno());
}

VARDECL: BTYPE VARDEFS ';' {
	$$ = new AST::VarDecl($1, $2, yyget_lineno());
}

VARDEFS: VARDEF {
	$$ = new AST::VarDefs(yyget_lineno());
	$$->list.push_back($1);
} | VARDEFS ',' VARDEF {
	$$ = $1;
	$$->list.push_back($3);
}

VARDEF: IDEXP ARR_INDEX {
	$$ = new AST::VarDef($1, $2, nullptr, yyget_lineno());
} | IDEXP ARR_INDEX '=' INITVAL {
	$$ = new AST::VarDef($1, $2, $4, yyget_lineno());
}

ARR_INDEX:  {
	$$ = new AST::Arr_Index(yyget_lineno());
} | ARR_INDEX '[' EXP ']' {
	$$ = $1;
	$$->list.push_back($3);
}

INITVAL: EXP {
	$$ = $1;
} | ARR_INIT {
	$$ = $1;
}

ARR_INIT: '{' INITVALS '}' {
	$$ = $2;
}

INITVALS: {
	$$ = new AST::InitVals(yyget_lineno());
} | INITVAL {
	$$ = new AST::InitVals($1, yyget_lineno());
} | INITVALS ',' INITVAL {
	$$ = $1;
	$$->list.push_back($3);
}

FUNCDEF: BTYPE IDEXP '(' PARAMETERS ')' BLOCK {
	$$ = new AST::FuncDef($1, $2, $4, $6, yyget_lineno());
}

PARAMETERS:  {
	$$ = new AST::Parameters(yyget_lineno());
} | PARAMETER {
	$$ = new AST::Parameters($1, yyget_lineno());
} | PARAMETERS ',' PARAMETER {
	$$ = $1;
	$$->list.push_back($3);
}

PARAMETER: BTYPE IDEXP {
	$$ = new AST::Parameter($1, $2, new AST::Arr_Index(yyget_lineno()), yyget_lineno());
} | BTYPE IDEXP '[' ']' ARR_INDEX {
	auto x = new AST::Arr_Index(yyget_lineno());
	x->list.push_back(new AST::IntNumber(1, yyget_lineno()));
	for (const auto &y:$5->list) {
		x->list.push_back(y);
	}
	$$ = new AST::Parameter($1, $2, x, yyget_lineno());
}

BLOCK: '{' BLOCKITEMS '}' {
	$$ = new AST::Block($2, yyget_lineno());
}

BLOCKITEMS:  {
	$$ = new AST::BlockItems(yyget_lineno());
} | BLOCKITEMS BLOCKITEM {
	$$ = $1;
	$$->list.push_back($2);
}

BLOCKITEM: DECL {
	$$ = $1;
} | STMT {
	$$ = $1;
}

STMT: ASSIGNSTMT ';' {
	$$ = $1;
} | ';' {
	$$ = new AST::ExpStmt(new AST::IntNumber(0, yyget_lineno()), yyget_lineno());
} | EXPSTMT ';' {
	$$ = $1;
} | BLOCK {
	$$ = $1;
} | IFSTMT {
	$$ = $1;
} | WHILESTMT {
	$$ = $1;
} | BREAKSTMT ';' {
	$$ = $1;
} | CONTINUESTMT ';' {
	$$ = $1;
} | RETURNSTMT ';' {
	$$ = $1;
} | STARTTIMESTMT ';' {
	$$ = $1;
} | STOPTIMESTMT ';' {
	$$ = $1;
}

ASSIGNSTMT: LVAL '=' EXP {
	$$ = new AST::AssignStmt($1, $3, yyget_lineno());
}

EXPSTMT: EXP {
	$$ = new AST::ExpStmt($1, yyget_lineno());
}

IFSTMT: IF '(' EXP ')' STMT %prec prec2 {
	$$ = new AST::IfStmt($3, $5, nullptr, yyget_lineno());
} | IF '(' EXP ')' STMT ELSE STMT {
	$$ = new AST::IfStmt($3, $5, $7, yyget_lineno());
}

WHILESTMT: WHILE '(' EXP ')' STMT {
	$$ = new AST::WhileStmt($3, $5, yyget_lineno());
}

BREAKSTMT: BREAK {
	$$ = new AST::BreakStmt(yyget_lineno());
}

CONTINUESTMT: CONTINUE {
	$$ = new AST::ContinueStmt(yyget_lineno());
}

RETURNSTMT: RETURN {
	$$ = new AST::ReturnStmt(nullptr, yyget_lineno());
} | RETURN EXP {
	$$ = new AST::ReturnStmt($2, yyget_lineno());
}

STARTTIMESTMT: STARTTIME {
	$$ = new AST::StarttimeStmt(yyget_lineno());
}

STOPTIMESTMT: STOPTIME {
	$$ = new AST::StoptimeStmt(yyget_lineno());
}

EXP: PRIMARYEXP {
	$$ = $1;
} | UNARYEXP {
	$$ = $1;
} | MULEXP {
	$$ = $1;
} | ADDEXP {
	$$ = $1;
} | RELEXP {
	$$ = $1;
} | EQEXP {
	$$ = $1;
} | LANDEXP {
	$$ = $1;
} | LOREXP {
	$$ = $1;
} | CALLEXP {
	$$ = $1;
} | FORMATEXP {
	$$ = $1;
}

IDEXP: ID {
	$$ = new AST::IdExp($1, yyget_lineno());
}

LVAL: IDEXP ARR_INDEX {
	$$ = new AST::Lval($1, $2, yyget_lineno());
}

PRIMARYEXP: '(' EXP ')' {
	$$ = $2;
} | LVAL {
	$$ = $1;
} | NUMBER {
	$$ = $1;
}

NUMBER: INTNUMBER {
	$$ = new AST::IntNumber($1, yyget_lineno());
} | FLOATNUMBER {
	$$ = new AST::FloatNumber($1, yyget_lineno());
}

UNARYEXP: UNARYOP EXP %prec prec1 {
	$$ = new AST::UnaryExp($1, $2, yyget_lineno());
} 

CALLEXP: IDEXP '(' EXPS ')' {
	$$ = new AST::CallExp($1, $3, yyget_lineno());
}


UNARYOP: ADDOP {
	$$ = AST::unary_op_t::ADD;
} | SUBOP {
	$$ = AST::unary_op_t::SUB;
} | NOTOP {
	$$ = AST::unary_op_t::NOT;
}

EXPS:  {
	$$ = new AST::Exps(yyget_lineno());
} | EXP {
	$$ = new AST::Exps($1, yyget_lineno());
} | EXPS ',' EXP {
	$$ = $1;
	$$->list.push_back($3);
}

MULEXP: EXP MULOP EXP {
	$$ = new AST::MulExp($1, $2, $3, yyget_lineno());
}

ADDEXP: EXP ADDOP EXP {
	$$ = new AST::AddExp($1, AST::add_op_t::ADD, $3, yyget_lineno());
} | EXP SUBOP EXP {
	$$ = new AST::AddExp($1, AST::add_op_t::SUB, $3, yyget_lineno());
}

RELEXP: EXP RELOP EXP {
	$$ = new AST::RelExp($1, $2, $3, yyget_lineno());
}

EQEXP: EXP EQUALOP EXP {
	$$ = new AST::EqExp($1, $2, $3, yyget_lineno());
}

LANDEXP: EXP LANDOP EXP {
	$$ = new AST::LAndExp($1, $3, yyget_lineno());
}

LOREXP: EXP LOROP EXP {
	$$ = new AST::LOrExp($1, $3, yyget_lineno());
}

FORMATEXP: FORMAT {
	$$ = new AST::FormatExp($1, yyget_lineno());
}

%%
