
#include <stdio.h>
#include <string>
#include <iostream>
#include <cassert>
#include "utils/table.hpp"
#include "structure/ast/ast.hpp"
#include "structure/mir/mir.hpp"
#include "utils/temp.hpp"
#include "backend/canon/canon.hpp"
#include "backend/cfg/bg.hpp"
#include "backend/optimizer/cp.hpp"
#include "backend/optimizer/de.hpp"
#include "backend/alloc/regalloc.hpp"
#include "frontend/lexer.hpp"
#include "frontend/parser.hpp"

using std::cout, std::endl;

// extern int yydebug;
extern int yyparse();
extern "C" FILE *yyin;
extern AST::CompUnits* root;

int main(int argc, char** argv) {
  //compiler testcase.sysy -S -o testcase.s [-O1]
  std::string source_code_path, target_code_path;
  bool is_user_defined_output = false, need_assem_output = false;
  for(int i = 0; i <argc; i++){
    std::string argv_s = argv[i];
    if(argv_s.find(".sy") != std::string::npos){
      if(source_code_path.empty()) source_code_path = argv_s;
    }else if(argv_s.find(".s") != std::string::npos){
      if(target_code_path.empty()) target_code_path = argv_s;
    }else if(argv_s == "-S" || argv_s == "-s"){
      need_assem_output = true;
    }else if(argv_s == "-O" || argv_s == "-o"){
      is_user_defined_output = true;
    }
  }
  if(!is_user_defined_output){
    target_code_path = source_code_path;
    target_code_path.replace(target_code_path.find(".sy"), 3, ".s");
  }

    
  std::cout << "-------------" << source_code_path << "-------------" << std::endl;
  yyin = fopen(source_code_path.c_str(), "r");
  assert(yyin);
  yyparse();
  fclose(yyin);
  // root->printTree(0);
  auto venv = new TABLE::SymbolTable<ENTRY::Entry*>();
  auto fenv = new TABLE::SymbolTable<ENTRY::EnFunction*>();
  mir::UnitList* newUnitList = root->semanticAnalysis(venv, fenv);
  for(auto& v: *venv) delete v.second.first;
  for(auto& f: *fenv) delete f.second.first;
  delete venv;
  delete fenv;
  //for (auto i: *newUnitList) {
    //i->print();
  //}

  auto ul = newUnitList;
  mir::get_global_table(ul);
  // cout << "get_global_table pass" << endl;
  auto tot_il = new assem::InstrList();
  

  for(auto& u: *ul){
    if(u->kind == mir::UnitType::FUN){
      auto fun = static_cast<mir::Fun*>(u);
      fun->stm = fun->stm->stm_deepcopy();
      fun->stm->fix(true);
      fun->stm->fix(false);
      //fun->print();
      fun->stm->flatten();
      //cout << "fix & flatten pass" << endl;
      // fun->print();
      auto sl = canon::linearize(fun->stm);
      auto b = canon::basicBlocks(sl, *fun->name);
      //cfg::BlockGraph* bg = new cfg::BlockGraph(b);
      //bg->show();
      //ssa::Ssa* ssa = new ssa::Ssa(bg->mynodes, fun->args);
      // for(auto node: *(ssa->nodelist)){
      //   auto stmlist = std::any_cast<mir::StmList*>(node->info);
      //   for(auto& stm: *stmlist){
      //     printf("stm");
      //     stm->print();
      //   }
      // }
      
      // cout << "back to ir pass"<<endl;
      // for(auto node: *(ssa->nodelist)){
      //   auto stmlist = std::any_cast<mir::StmList*>(node->info);
      //   for(auto& stm: *stmlist){
      //     printf("after ssa");
      //     stm->print();
      //   }
      // }
      //ssa::ConstPropagation* cp = new ssa::ConstPropagation(bg->mynodes);
      // // for(auto node: *(cp->nodelist)){
      // //   auto stmlist = std::any_cast<mir::StmList*>(node->info);
      // //   for(auto& stm: *stmlist){
      // //     printf("cp:");
      // //     stm->print();
      // //   }
      // // }
      //ssa::test_cdg();
      //ssa::DeadcodeElimination* de = new ssa::DeadcodeElimination(bg->mynodes);
      // // for(auto node: *(de->nodelist)){
      // //   auto stmlist = std::any_cast<mir::StmList*>(node->info);
      // //   for(auto& stm: *stmlist){
      // //     printf("de:");
      // //     stm->print();
      // //   }
      // // }
      //ssa->back_to_ir();
      auto sl_after_trace = canon::traceSchedule(b);
      //for(auto& s: *sl_after_trace) s->print();
      // for(auto& s: *sl_after_trace) s->print();
      auto il = mir::ir2assem(fun, sl_after_trace);
      //for(auto& i: *il) i->print();
      auto alloc = alloc::Allocator();
      alloc.reg_alloc(il);
      // for(auto& i: *il) i->print();

      tot_il->splice(tot_il->end(), *il);
     
      
    }
  }
  //for(auto& i: *tot_il) i->print();
  
  tot_il = mir::assem_add_epilog(tot_il);
  
  FILE* output = freopen(target_code_path.c_str(), "w", stdout);
  for(auto& i: *tot_il) i->print();
  fclose(output);
      
	return 0;
}
