#include "mir.hpp"
#include <string>
#include <cassert>

namespace mir{
using namespace std;

void Fun::ir2llvm(assem::Instrlist* instrlist){
    std::string assem = "define i64 @" + *name;
    assem += "(";
    for(auto& arg: *args){
        assem += *arg->id;
    }
    assem += "){";
    instrlist->emplace_back(
        new assem::OperInstr(assem, new utils::TempList(), new utils::TempList(), new utils::LabelList())
    );
}
void Seq::ir2llvm(assem::Instrlist* instrlist){
    left->ir2llvm(instrlist);   
    right->ir2llvm(instrlist);   
}
void Label::ir2llvm(assem::Instrlist* instrlist){
    instrlist->emplace_back(
        new assem::LabelInstr(label)
    );
}
void Jump::ir2llvm(assem::Instrlist* instrlist){
    instrlist->emplace_back(
        new assem::OperInstr("br label " + *label->id, new utils::TempList(), new utils::TempList(), new utils::LabelList{label})
    );
}
void Cjump::ir2llvm(assem::Instrlist* instrlist){
    utils::Temp* l = left->ir2llvm(instrlist);
    utils::Temp* r = right->ir2llvm(instrlist);
    utils::Temp* icmp_dst = new utils::Temp();
    instrlist->emplace_back(
        new assem::OperInstr(
            "`d0 = icmp " + relop2str(op) + "i64 `s0, `s1",
            new utils::TempList{l, r},
            new utils::TempList{icmp_dst},
            new utils::LabelList()
        )
    );
    instrlist->emplace_back(
        new assem::OperInstr(
            "br i1 `s0, label " + *true_label->id + ", label " + *false_label->id,
            new utils::TempList{icmp_dst},
            new utils::TempList(),
            new utils::LabelList{true_label, false_label}
        )
    );
}
void Move::ir2llvm(assem::Instrlist* instrlist){
    utils::Temp* r = src->ir2llvm(instrlist);
    if(dst->kind == ExpType::MEM){
        Mem* mem = static_cast<Mem*>(dst);
        utils::Temp* mem_tmp = mem->exp->ir2llvm(instrlist);
        utils::Temp* mem_dst = new utils::Temp();
        instrlist->emplace_back(
            new assem::OperInstr(
                "`d0 = inttoptr i64 `s0 to ptr",
                new utils::TempList{mem_tmp},
                new utils::TempList{mem_dst},
                new utils::LabelList()
            )
        );
        instrlist->emplace_back(
            new assem::OperInstr(
                "store i64 `s0, ptr `s1", 
                new utils::TempList{r, mem_dst},
                new utils::TempList(),
                new utils::LabelList()
            )
        );
    }else{
        utils::Temp* l = dst->ir2llvm(instrlist);
        instrlist->emplace_back(
            new assem::MoveInstr(
                "`d0 = add i64 `s0, 0",
                new utils::TempList{r},
                new utils::TempList{l}
            )
        );
    }
    
}
void ExpStm::ir2llvm(assem::Instrlist* instrlist){
    assert(exp->kind == ExpType::CALL);
    Call* call = static_cast<Call*>(exp);
    string assem = "call void " + *call->name;
    utils::TempList* src_tl = new utils::TempList();
    assem += "(";
    int param_id = 0;
    for(auto& arg: *call->args){
        utils::Temp* t = arg->ir2llvm(instrlist);
        src_tl->emplace_back(t);
        assem += "`s" + to_string(param_id++);
    }
    assem += ")";
    instrlist->emplace_back(
        new assem::OperInstr(
            assem, src_tl, new utils::TempList(), new utils::LabelList()
        )
    );
}
void Return::ir2llvm(assem::Instrlist* instrlist){
    utils::Temp* ret = exp->ir2llvm(instrlist);
    instrlist->emplace_back(
        new assem::OperInstr(
            "ret i64 `s0",
            new utils::TempList{ret},
            new utils::TempList(),
            new utils::LabelList()
        )
    );
}
utils::Temp* Phi::ir2llvm(assem::Instrlist* instrlist){

}
utils::Temp* Binop::ir2llvm(assem::Instrlist* instrlist){
    utils::Temp* l = left->ir2llvm(instrlist);
    utils::Temp* r = right->ir2llvm(instrlist);
    utils::Temp* result = new utils::Temp();
    instrlist->emplace_back(
        new assem::OperInstr(
            "`d0 = " + binop2str(op) + "i64 `s0, `s1",
            new utils::TempList{l, r},
            new utils::TempList{result},
            new utils::LabelList()
        )
    );
    return result;
}
utils::Temp* Mem::ir2llvm(assem::Instrlist* instrlist){
    utils::Temp* mem_tmp = exp->ir2llvm(instrlist);
    utils::Temp* mem_ptr = new utils::Temp();
    utils::Temp* mem_dst = new utils::Temp();
    instrlist->emplace_back(
        new assem::OperInstr(
            "`d0 = inttoptr i64 `s0 to ptr",
            new utils::TempList{mem_tmp},
            new utils::TempList{mem_ptr},
            new utils::LabelList()
        )
    );
    instrlist->emplace_back(
        new assem::OperInstr(
            "`d0 = load i64, ptr `s0",
            new utils::TempList{mem_ptr},
            new utils::TempList{mem_dst},
            new utils::LabelList()
        )
    );
    return mem_dst;
}
utils::Temp* Temp::ir2llvm(assem::Instrlist* instrlist){
    return temp;
}
utils::Temp* Eseq::ir2llvm(assem::Instrlist* instrlist){
    stm->ir2llvm(instrlist);
    utils::Temp* exp_tmp = exp->ir2llvm(instrlist);
    return exp_tmp;
}
utils::Temp* Name::ir2llvm(assem::Instrlist* instrlist){
    assert(0);
}
utils::Temp* Format::ir2llvm(assem::Instrlist* instrlist){
    assert(0);
}
utils::Temp* Const::ir2llvm(assem::Instrlist* instrlist){
    utils::Temp* const_tmp = new utils::Temp();
    instrlist->emplace_back(
        new assem::OperInstr(
            "`d0 = add i64 0, " + this->to_string(),
            new utils::TempList(),
            new utils::TempList{const_tmp},
            new utils::LabelList()
        )
    );
    return const_tmp;
}
utils::Temp* Call::ir2llvm(assem::Instrlist* instrlist){
    
    string assem = "`d0 = call i64 " + *name;
    utils::TempList* src_tl = new utils::TempList();
    utils::TempList* dst_tl = new utils::TempList{new utils::Temp()};
    assem += "(";
    int param_id = 0;
    for(auto& arg: *args){
        utils::Temp* t = arg->ir2llvm(instrlist);
        src_tl->emplace_back(t);
        assem += "`s" + to_string(param_id++);
    }
    assem += ")";
    instrlist->emplace_back(
        new assem::OperInstr(
            assem, src_tl, dst_tl, new utils::LabelList()
        )
    );
    return dst_tl->front();
}

}