#ifndef __TABLE_HPP
#define __TABLE_HPP

#include <iostream>
#include <string>
#include <unordered_map>
#include <deque>


namespace TABLE {

template <typename T_key, typename T_value> class table {
private:
    std::unordered_map<T_key, T_value> map;
    std::deque<std::pair<T_key, std::pair<T_value, bool>>> deq;

public:
    void enter(T_key, T_value);
    bool exist(T_key);
    T_value look(T_key);
    std::pair<T_key, T_value> pop();
    auto begin() -> decltype(deq.begin()) { return deq.begin(); }
    auto end() -> decltype(deq.end()) { return deq.end(); }
};


template <typename T_value> class SymbolTable {
private:
    TABLE::table<std::string, T_value> table;

public:
    SymbolTable() {}
    ~SymbolTable() {}

    void enter(std::string, T_value);
    T_value look(std::string);
    bool exist(std::string);
    void beginScope(T_value useless);  // anything
    void endScope();
    std::pair<std::string, T_value> pop();
    auto begin() -> decltype(table.begin()) { return table.begin(); }
    auto end() -> decltype(table.end()) { return table.end(); }
};


}

template <typename T_key, typename T_value>
void TABLE::table<T_key, T_value>::enter(T_key key, T_value value) {
    if (this->map.count(key))
        this->deq.push_back(std::make_pair(key, std::make_pair(map[key], true)));
    else
        this->deq.push_back(std::make_pair(key, std::make_pair(map[key], false)));
    map[key] = value;
}

template <typename T_key, typename T_value> bool TABLE::table<T_key, T_value>::exist(T_key key) {
    return map.find(key) != map.end();
}

template <typename T_key, typename T_value> T_value TABLE::table<T_key, T_value>::look(T_key key) {
    return map.find(key)->second;
}

template <typename T_key, typename T_value>
std::pair<T_key, T_value> TABLE::table<T_key, T_value>::pop() {
    assert(deq.size());
    std::pair<T_key, T_value> ret = *map.find(deq.back().first);
    if (deq.back().second.second)
        map[deq.back().first] = deq.back().second.first;
    else
        map.erase(deq.back().first);
    deq.pop_back();

    return ret;
}

template <typename T_value> void TABLE::SymbolTable<T_value>::enter(std::string str, T_value value) {
    this->table.enter(str, value);
}

template <typename T_value> bool TABLE::SymbolTable<T_value>::exist(std::string str) {
    return this->table.exist(str);
}

template <typename T_value> T_value TABLE::SymbolTable<T_value>::look(std::string str) {
    return this->table.look(str);
}

template <typename T_value> void TABLE::SymbolTable<T_value>::beginScope(T_value useless) {
    this->enter(std::string("[_A__SCOPE_]"), useless);
}

template <typename T_value> void TABLE::SymbolTable<T_value>::endScope() {
    while ((this->pop().first) != std::string("[_A__SCOPE_]"))
        ;
}

template <typename T_value> std::pair<std::string, T_value> TABLE::SymbolTable<T_value>::pop() {
    return this->table.pop();
}

#endif // __TABLE_H
